#include "theBiasPlanet/coreUtilities/pipes/ObjectsPipe.hpp"
#include <chrono>
#include <exception>
#include "theBiasPlanet/coreUtilities/inputsHandling/NoMoreDataException.hpp"
#include "theBiasPlanet/coreUtilities/inputsHandling/NoMoreNeedsException.hpp"
#include "theBiasPlanet/coreUtilities/messagingHandling/Publisher.hpp"
#include "theBiasPlanet/coreUtilities/timersHandling/TimeOutException.hpp"

using namespace ::std::chrono;
using namespace ::theBiasPlanet::coreUtilities::inputsHandling;
using namespace ::theBiasPlanet::coreUtilities::messagingHandling;
using namespace ::theBiasPlanet::coreUtilities::timersHandling;

namespace theBiasPlanet {
	namespace coreUtilities {
		namespace pipes {
			template <typename T> ObjectsPipe <T>::ObjectsPipe (int const & a_bufferSize, bool const & a_notificationIsDelayed): i_objects (new T [a_bufferSize]), i_bufferSize (a_bufferSize), i_notificationIsDelayed (a_notificationIsDelayed) {
			}
			
			template <typename T> ObjectsPipe <T>::~ObjectsPipe () {
				delete i_objects;
			}
			
			template <typename T> bool ObjectsPipe <T>::isEmptyWithoutLocking () {
				return i_dataStartIndex == GeneralConstantsConstantsGroup::c_iterationStartNumber && i_dataUntilIndex == GeneralConstantsConstantsGroup::c_iterationStartNumber;
			}
			
			template <typename T> bool ObjectsPipe <T>::isFullWithoutLocking () {
				return (i_dataStartIndex == GeneralConstantsConstantsGroup::c_iterationStartNumber && i_dataUntilIndex == i_bufferSize) || (i_dataStartIndex != GeneralConstantsConstantsGroup::c_iterationStartNumber && i_dataStartIndex == i_dataUntilIndex);
			}
			
			template <typename T> void ObjectsPipe <T>::writeWithoutLocking (T const & a_object, unique_lock <recursive_mutex> * a_lock, long const & a_timeOutPeriodInMilliseconds) {
				if (i_isFinishedReading) {
					throw NoMoreNeedsException ("");
				}
				if (i_isFinishedWriting) {
					i_isFinishedWriting = false;
				}
				while (true) {
					if (isFullWithoutLocking ()) {
						try {
							if (a_timeOutPeriodInMilliseconds == -1) {
								i_threadCondition.wait (*a_lock);
							}
							else if (a_timeOutPeriodInMilliseconds == 0) {
							}
							else {
								i_threadCondition.wait_for (*a_lock, milliseconds (a_timeOutPeriodInMilliseconds));
							}
						}
						catch (exception & l_exception) {
							Publisher::logErrorInformation (l_exception);
						}
					}
					// Checked again because the status may have changed while this thread was waiting.
					if (i_isFinishedReading) {
						throw NoMoreNeedsException ("");
					}
					if (!isFullWithoutLocking ()) {
						bool l_wasEmpty = isEmptyWithoutLocking ();
						if (i_dataUntilIndex == i_bufferSize) {
							i_objects [GeneralConstantsConstantsGroup::c_iterationStartNumber] = a_object;
							i_dataUntilIndex = GeneralConstantsConstantsGroup::c_iterationStartNumber + 1;
						}
						else {
							i_objects [i_dataUntilIndex] = a_object;
							i_dataUntilIndex ++;
						}
						if ( (!i_notificationIsDelayed && l_wasEmpty) || (i_notificationIsDelayed && isFullWithoutLocking ())) {
							i_threadCondition.notify_all ();
						}
						return;
					}
					else {
						if (a_timeOutPeriodInMilliseconds != -1) {
							throw TimeOutException ("");
						}
					}
				}
			}
			
			template <typename T> T ObjectsPipe <T>::readWithoutLocking (unique_lock <recursive_mutex> * a_lock, long const & a_timeOutPeriodInMilliseconds) {
				T l_readObject;
				if (i_isFinishedReading) {
					i_isFinishedReading = false;
				}
				while (true) {
					if (isEmptyWithoutLocking ()) {
						if (!i_isFinishedWriting) {
							try {
								if (a_timeOutPeriodInMilliseconds == -1) {
									i_threadCondition.wait (*a_lock);
								}
								else if (a_timeOutPeriodInMilliseconds == 0) {
								}
								else {
									i_threadCondition.wait_for (*a_lock, milliseconds (a_timeOutPeriodInMilliseconds));
								}
							}
							catch (exception & l_exception) {
								Publisher::logErrorInformation (l_exception);
							}
						}
						else {
							throw NoMoreDataException ("");
						}
					}
					// Checked again because the status may have changed while this thread was waiting.
					if (!isEmptyWithoutLocking ()) {
						bool l_wasFull = isFullWithoutLocking ();
						l_readObject = i_objects [i_dataStartIndex];
						// 'i_objects [i_dataStartIndex]' is left although is invalidated.
						i_dataStartIndex ++;
						if (i_dataStartIndex == i_dataUntilIndex) {
							i_dataStartIndex = GeneralConstantsConstantsGroup::c_iterationStartNumber;
							i_dataUntilIndex = GeneralConstantsConstantsGroup::c_iterationStartNumber;
						}
						else {
							if (i_dataStartIndex == i_bufferSize) {
								i_dataStartIndex = GeneralConstantsConstantsGroup::c_iterationStartNumber;
							}
						}
						if ( (!i_notificationIsDelayed && l_wasFull) || (i_notificationIsDelayed && isEmptyWithoutLocking ())) {
							i_threadCondition.notify_all ();
						}
						return l_readObject;
					}
					else {
						if (i_isFinishedWriting) {
							throw NoMoreDataException ("");
						}
						if (a_timeOutPeriodInMilliseconds != -1) {
							throw TimeOutException ("");
						}
					}
				}
			}
			
			template <typename T> bool ObjectsPipe <T>::isEmpty () {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				return isEmptyWithoutLocking ();
			}
			
			template <typename T> bool ObjectsPipe <T>::isFull () {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				return isFullWithoutLocking ();
			}
			
			template <typename T> void ObjectsPipe <T>::write (T const & a_object, long const & a_timeOutPeriodInMilliseconds) {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				writeWithoutLocking (a_object, &l_lock, a_timeOutPeriodInMilliseconds);
			}
			
			template <typename T> int ObjectsPipe <T>::write (T const * const a_objects, int const & a_offset, int const & a_length, long const & a_timeOutPeriodInMilliseconds) {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				int l_writtenLength = 0;
				for (l_writtenLength = 0; l_writtenLength < a_length; l_writtenLength ++) {
					try {
						if (l_writtenLength == 0 || ! (isFullWithoutLocking ())) {
							writeWithoutLocking (a_objects [a_offset + l_writtenLength], &l_lock, a_timeOutPeriodInMilliseconds);
						}
					}
					catch (NoMoreNeedsException & l_exception) {
						if (l_writtenLength == 0) {
							throw l_exception;
						}
						else {
							break;
						}
					}
				}
				return l_writtenLength;
			}
			
			template <typename T> T ObjectsPipe <T>::read (long const & a_timeOutPeriodInMilliseconds) {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				return readWithoutLocking (&l_lock, a_timeOutPeriodInMilliseconds);
			}
			
			template <typename T> int ObjectsPipe <T>::read (T * const a_objects, int const & a_offset, int const & a_length, long const & a_timeOutPeriodInMilliseconds) {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				int l_readLength = 0;
				for (; l_readLength < a_length; l_readLength ++) {
					if ( (l_readLength == 0) || ! (isEmptyWithoutLocking ())) {
						a_objects [a_offset + l_readLength] = readWithoutLocking (&l_lock, a_timeOutPeriodInMilliseconds);
					}
					else {
						break;
					}
				}
				return l_readLength;
			}
			
			template <typename T> list <T> ObjectsPipe <T>::readWholeData () {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				list <T> l_objectsList;
				while (true) {
					try {
						l_objectsList.push_back (readWithoutLocking (&l_lock));
					}
					catch (NoMoreDataException & l_exception) {
						break;
					}
				}
				return l_objectsList;
			}
			
			template <typename T> void ObjectsPipe <T>::finishWriting () {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				i_isFinishedWriting = true;
				i_threadCondition.notify_all ();
			}
			
			template <typename T> void ObjectsPipe <T>::finishReading () {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				i_isFinishedReading = true;
				i_threadCondition.notify_all ();
			}
			
			template <typename T> void ObjectsPipe <T>::reset () {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				i_isFinishedWriting = false;
				i_isFinishedReading = false;
				i_dataStartIndex = GeneralConstantsConstantsGroup::c_iterationStartNumber;
				i_dataUntilIndex = GeneralConstantsConstantsGroup::c_iterationStartNumber;
			}
		}
	}
}

